import re
from itertools import count
from urllib.parse import quote as url_quote

from django.core.mail import EmailMultiAlternatives, EmailMessage
from django.template import Context
from django.urls import reverse

from nuntius import app_settings
from nuntius.utils.messages import sign_url, extend_query

RE_URL = re.compile(
    r"(?P<prefix><a[^>]* href\s*=[\s\"']*)(?P<url>http[^\"'>\s]+)",
    flags=re.MULTILINE | re.IGNORECASE,
)


def insert_tracking_image_template(html_message):
    # use old style formatting to avoid ugly escaping of template variable
    img_url = "%s%sopen/{{ nuntius_tracking_id }}" % (
        app_settings.PUBLIC_URL,
        reverse("nuntius_mount_path"),
    )
    img = f'<img src="{img_url}" width="1" height="1" alt="nt">'
    return re.sub(
        r"(</body\b)", img + r"\1", html_message, flags=re.MULTILINE | re.IGNORECASE
    )


def make_tracking_url(url, campaign, tracking_id, link_index):
    url = extend_query(
        url,
        defaults={
            "utm_content": f"link-{link_index}",
            "utm_term": getattr(campaign.segment, "utm_term", ""),
        },
    )

    relative_url = reverse(
        "nuntius_track_click",
        kwargs={
            "tracking_id": tracking_id,
            "signature": sign_url(campaign, url),
            "link": url_quote(url, safe=""),
        },
    )
    return f"{app_settings.LINKS_URL}{relative_url}"


def href_url_replacer(campaign, tracking_id):
    link_counter = count()

    def url_replace(match):
        url = make_tracking_url(
            match.group("url"), campaign, tracking_id, next(link_counter)
        )
        return f"{match.group('prefix')}{url}"

    return url_replace


def add_tracking_information(html_body: str, campaign, tracking_id):
    return RE_URL.sub(
        href_url_replacer(campaign=campaign, tracking_id=tracking_id), html_body
    )


def message_for_event(sent_event):
    """Generate an email message corresponding to a CampaignSentEvent instance

    :param sent_event: the CampaignSentEvent instance associating a campaign and a subscriber for which the
        message should be generated
    :type sent_event: class:`nuntius.models.CampaignSentEvent`
    :return: a complete email message, for the campaign, with the subscriber information interpolated
    :rtype: class:`django.core.mail.message.EmailMessage`
    """
    subscriber = sent_event.subscriber
    campaign = sent_event.campaign
    email = sent_event.email

    subscriber_data = Context(
        {
            "nuntius_tracking_id": sent_event.tracking_id,
            **subscriber.get_subscriber_data(),
        }
    )

    html_body = add_tracking_information(
        campaign.html_template.render(context=subscriber_data),
        campaign,
        sent_event.tracking_id,
    )
    text_body = campaign.text_template.render(context=subscriber_data)

    message_class = (
        EmailMultiAlternatives if (text_body and html_body) else EmailMessage
    )

    message = message_class(
        subject=campaign.message_subject,
        body=text_body or html_body,
        from_email=campaign.from_header,
        to=[email],
        reply_to=[campaign.reply_to_header]
        if campaign.reply_to_header is not None
        else None,
    )

    if text_body and html_body:
        message.attach_alternative(html_body, "text/html")
    elif html_body:
        message.content_subtype = "html"

    return message



from base64 import b64decode
from urllib.parse import urlparse, unquote
from urllib.request import urlopen

from PIL import Image
from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.db.models import F
from django.http import HttpResponse, HttpResponseBadRequest
from django.http.request import validate_host
from django.shortcuts import redirect, get_object_or_404
from django.views.decorators.cache import cache_control

from nuntius.models import MosaicoImage, CampaignSentEvent
from nuntius.utils.messages import (
    generate_placeholder,
    url_signature_is_valid,
    extend_query,
)

TRACKING_IMAGE_CONTENT = b64decode(
    "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNkYAAAAAYAAjCB0C8AAAAASUVORK5CYII="
)


@cache_control(public=True, max_age=3600)
def mosaico_image_processor_view(request):
    params = request.GET.get("params", "").split(",")

    if len(params) != 2:
        return HttpResponseBadRequest()

    try:
        (width, height) = (
            int(params[0].replace("null", "0")),
            int(params[1].replace("null", "0")),
        )
    except ValueError:
        return HttpResponseBadRequest()

    if request.GET.get("method") == "placeholder" and width and height:
        image = generate_placeholder(width, height)
        response = HttpResponse(content_type="image/png")
        image.save(response, "PNG")
        return response

    if request.GET.get("src") and (width or height):
        try:
            path = urlparse(unquote(request.GET.get("src"))).path.replace(
                settings.MEDIA_URL, "", 1
            )
        except ValueError:
            return HttpResponseBadRequest()

        image = get_object_or_404(MosaicoImage, file=path)
        try:
            image = Image.open(image.file.path)
        except NotImplementedError:
            image = Image.open(urlopen(image.file.url))

        if width and height:
            ratio = min(width / image.size[0], height / image.size[1])
        elif width:
            ratio = width / image.size[0]
        elif height:
            ratio = height / image.size[1]

        image.resize((round(size * ratio) for size in image.size), Image.ANTIALIAS)
        response = HttpResponse(content_type=f"image/{image.format.lower()}")
        image.save(response, image.format)
        return response

    return HttpResponseBadRequest()


def track_open_view(request, tracking_id):
    CampaignSentEvent.objects.filter(tracking_id=tracking_id).update(
        open_count=F("open_count") + 1
    )
    return HttpResponse(TRACKING_IMAGE_CONTENT, content_type="image/png")


def track_click_view(request, tracking_id, link, signature):
    campaign_sent_event = get_object_or_404(CampaignSentEvent, tracking_id=tracking_id)

    CampaignSentEvent.objects.filter(tracking_id=tracking_id).update(
        click_count=F("click_count") + 1
    )

    url = unquote(link)

    if not url_signature_is_valid(campaign_sent_event.campaign, url, signature):
        raise PermissionDenied()

    utm_campaign = campaign_sent_event.campaign.utm_name
    url = extend_query(
        url,
        defaults={"utm_campaign": utm_campaign},
        replace={"utm_source": "nuntius", "utm_medium": "email"},
    )
    return redirect(url)



import contextlib
import logging
import multiprocessing as mp
import multiprocessing.connection as mpc
import signal
import smtplib
from typing import Dict, List, Tuple

from django.core import mail
from django.core.mail import EmailMessage
from django.core.management import BaseCommand
from django.db import connection
from django.db.models import Exists, OuterRef
from tenacity import (
    retry,
    stop_after_attempt,
    retry_if_exception_type,
    wait_random_exponential,
    TryAgain,
)
from django.utils.translation import gettext as _, gettext_lazy

from nuntius import app_settings
from nuntius.messages import message_for_event
from nuntius.models import (
    Campaign,
    CampaignSentEvent,
    CampaignSentStatusType,
    AbstractSubscriber,
)
from nuntius.utils.processes import (
    gracefully_exit,
    print_stack_trace,
    reset_sigmask,
    RateLimiter,
    TokenBucket,
    RateMeter,
    GracefulExit,
    get_from_queue_or_quit,
    put_in_queue_or_quit,
    unexpected_exc_logger,
)

try:
    from anymail import exceptions as anymail_exceptions

    AnymailError = anymail_exceptions.AnymailError
    AnymailAPIError = anymail_exceptions.AnymailAPIError
    AnymailRecipientsRefused = anymail_exceptions.AnymailRecipientsRefused
except:

    class FakeException(Exception):
        pass

    AnymailError = AnymailRecipientsRefused = AnymailAPIError = FakeException


logger = logging.getLogger(__name__)


class ConnectionManager:
    """
    Manager around the SMTP or Mail API connection that handles reconnection and quitting
    """

    CONNECTION_ERRORS = (ConnectionError, smtplib.SMTPException, AnymailError)
    MAIL_SENDING_ERRORS = (
        smtplib.SMTPSenderRefused,
        smtplib.SMTPDataError,
        AnymailAPIError,
    )

    def __init__(self, quit_event):
        self._connection = mail.get_connection(backend=app_settings.EMAIL_BACKEND)
        self._message_counter = 0
        self._quit_event = quit_event

    @retry(
        wait=wait_random_exponential(max=30),
        retry=retry_if_exception_type(CONNECTION_ERRORS),
    )
    def open_connection(self):
        """
        Connect to the SMTP or API server and retry with exponential backoff.

        This function will try up to 5 times to connect to the server, allowing for shaky
        connections.
        """
        if self._quit_event.is_set():
            raise GracefulExit()

        try:
            self._connection.open()
        except Exception:
            self._connection.close()
            raise
        else:
            self._message_counter = 0

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_random_exponential(),
        retry=retry_if_exception_type(MAIL_SENDING_ERRORS),
    )
    def send_message(self, message: EmailMessage):
        """
        Send an email message and retry in cases of failures.

        This function will try sending up to 5 times, and uses an exponential
        backoff strategy, where a random waiting time is chosen between 0 and a max
        value that doubles every retry, to avoid all sender processes retrying at the
        same time.

        :param message: the email message to send
        :param connection: the connection to use to send the message
        """
        if self._quit_event.is_set():
            raise GracefulExit()

        if (
            app_settings.MAX_MESSAGES_PER_CONNECTION
            and self._message_counter >= app_settings.MAX_MESSAGES_PER_CONNECTION
        ):
            self._connection.close()
            self.open_connection()
        try:
            self._connection.send_messages([message])
            self._message_counter += 1
        except smtplib.SMTPServerDisconnected:
            self._connection.close()
            self.open_connection()
            raise TryAgain

    def __enter__(self):
        self.open_connection()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._connection.close()


@reset_sigmask
@unexpected_exc_logger
def sender_process(
    *,
    queue: mp.Queue,
    error_channel: mpc.Connection,
    quit_event: mp.Event,
    rate_limiter: RateLimiter = None,
    rate_meter: RateMeter = None,
):
    """
    Main function of the processes responsible for sending email messages.

    This process pulls `(message: EmailMessage, event: CampaignSentEvent)`
    tuples from the work queue, tries to send the message and saves the result
    on the `event`.

    Whenever an unexpected error happens (i.e. which does not seem to be linked
    to a particular recipient), the process signals the error on the
    `error_channel` by sending the campaign id.

    It monitors its own connection to the mail service, and resets it every
    `nuntius.app_settings.MAX_MESSAGES_PER_CONNECTION` messages.

    :param queue: The work queue on which tuples (EmailMessage, CampaignSentEvent) are received.
    :type queue: class:`multiprocessing.Queue`

    :param error_channel: channel on which campaign ids of failing campaigns are sent.
    :type error_channel: class:`multiprocessing.connection.Connection`

    :param quit_event: event that may be used by the main process to signal to senders they need to quit
    :type quit_event: class:`multiprocessing.Event`

    :param rate_limiter: rate limiter to limit the rate of message sendings.
    :type rate_limiter: class:`nuntius.utils.processes.RateLimiter`

    :param rate_meter: rate meter to allow measuring the sending speed
    :type rate_meter: class:`nuntius.utils.processes.RateMeter`
    """
    message: EmailMessage
    sent_event_id: int

    try:
        with ConnectionManager(quit_event) as connection_manager:
            while True:
                # the timeout allows the loop to start again every few seconds so that the
                # quit_event is checked and the process can quit if it has to.
                message, sent_event_id = get_from_queue_or_quit(
                    queue,
                    event=quit_event,
                    polling_period=app_settings.POLLING_INTERVAL,
                )

                sent_event_qs = CampaignSentEvent.objects.filter(id=sent_event_id)
                # messages generated by the campaign manager should only have one recipient
                email = message.to[0]

                # rate limit just before sending
                if rate_limiter:
                    rate_limiter.take()

                try:
                    connection_manager.send_message(message)
                except (smtplib.SMTPRecipientsRefused, AnymailRecipientsRefused):
                    # exceptions linked to a specific recipient need not stop the sending
                    CampaignSentEvent.objects.filter(id=sent_event_id).update(
                        result=CampaignSentStatusType.BLOCKED
                    )
                except GracefulExit:
                    raise
                except Exception:
                    campaign = Campaign.objects.get(campaignsentevent__id=sent_event_id)
                    error_channel.send(campaign.id)
                    logger.error(
                        _("Error while sending email for campaign %(campaign)s")
                        % repr(campaign),
                        exc_info=True,
                    )

                else:
                    if rate_meter:
                        rate_meter.count_up()
                    if hasattr(message, "anymail_status"):
                        if message.anymail_status.recipients[email].status in [
                            "invalid",
                            "rejected",
                            "failed",
                        ]:
                            sent_event_qs.update(result=CampaignSentStatusType.REJECTED)
                        else:
                            sent_event_qs.update(
                                result=CampaignSentStatusType.UNKNOWN,
                                esp_message_id=message.anymail_status.recipients[
                                    email
                                ].message_id,
                            )
                    else:
                        sent_event_qs.update(result=CampaignSentStatusType.UNKNOWN)
    except GracefulExit:
        return


@reset_sigmask
@unexpected_exc_logger
def campaign_manager_process(
    *, campaign: Campaign, queue: mp.Queue, quit_event: mp.Event
):
    """
    Main function of the process responsible for scheduling the sending of campaigns

    :param campaign: the campaign for which messages must be sent
    :type campaign: :class:`nuntius.models.Campaign`

    :param queue: the work queue on which email messages are put
    :type queue: : class:`multiprocessing.Queue`

    :param quit_event: an event that may be used by the main process to tell the manager it needs to quit
    :type quit_event: class:`multiprocessing.Event`
    """
    queryset = campaign.get_subscribers_queryset()
    # eliminate people who already received the message
    queryset = queryset.annotate(
        already_sent=Exists(
            CampaignSentEvent.objects.filter(
                subscriber_id=OuterRef("pk"), campaign_id=campaign.id
            ).exclude(result=CampaignSentStatusType.PENDING)
        )
    ).filter(already_sent=False)

    campaign_finished = False

    for subscriber in queryset.iterator():
        if quit_event.is_set():
            break

        if subscriber.get_subscriber_status() != AbstractSubscriber.STATUS_SUBSCRIBED:
            continue

        sent_event = campaign.get_event_for_subscriber(subscriber)

        if sent_event.result != CampaignSentStatusType.PENDING:
            # just in case there is another nuntius_worker started, but this should not happen
            continue

        message = message_for_event(sent_event)

        try:
            put_in_queue_or_quit(
                queue,
                (message, sent_event.id),
                event=quit_event,
                polling_period=app_settings.POLLING_INTERVAL,
            )
        except GracefulExit:
            break
    else:
        campaign_finished = True

    queue.close()
    queue.join_thread()
    # everything has been scheduled for sending:
    if campaign_finished:
        campaign.status = Campaign.STATUS_SENT
        campaign.save()


class Command(BaseCommand):
    STATS_MESSAGE = gettext_lazy(
        """STATISTICS
Message queue size: %(queue_size)s
Sender processes: %(sender_processes)s
Campaign managers: %(campaign_managers)s
Token bucket current capacity: %(bucket_capacity)s
Current sending rate: %(sending_rate)s"""
    )

    def handle(self, *args, **options):
        # used by campaign managers processes to queue emails to send
        self.queue = mp.Queue(maxsize=app_settings.MAX_CONCURRENT_SENDERS * 2)

        # used to signals to sender they should exist once the queue has been emptied
        self.senders_quit_event = mp.Event()

        # used by email senders to make sure they're not going over the max rate
        self.rate_limiter = TokenBucket(
            max=app_settings.MAX_CONCURRENT_SENDERS * 2,
            rate=app_settings.MAX_SENDING_RATE,
        )

        # allow us to measure the sending speed
        self.rate_meter = RateMeter(0.3, 0.5)

        # used by the main process to monitor the sender processes
        self.sender_processes: List[mp.Process] = []

        # used by sender processes to signal when a campaign failed, so that it can be stopped by the main process
        self.sender_pipes: List[mp.connection.Connection] = []

        # used by the main process to monitor campaign managers and tell them to quit
        self.campaign_manager_processes: Dict[int, Tuple[mp.Process, mp.Event]] = {}

        self._setup_signals()

        self.run_loop()

    def _setup_signals(self):
        signal.signal(signal.SIGINT, signal.default_int_handler)
        signal.signal(signal.SIGTERM, gracefully_exit)
        signal.signal(signal.SIGUSR1, self.print_stats)
        signal.signal(signal.SIGUSR2, print_stack_trace)

    @contextlib.contextmanager
    def _setup_signal_handlers_for_children(self):
        try:
            signal.pthread_sigmask(signal.SIG_BLOCK, [signal.SIGTERM, signal.SIGINT])
            signal.signal(signal.SIGINT, signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_DFL)
            signal.signal(signal.SIGUSR1, signal.SIG_DFL)
            signal.signal(signal.SIGUSR2, signal.SIG_DFL)
            yield
        finally:
            self._setup_signals()
            signal.pthread_sigmask(signal.SIG_UNBLOCK, [signal.SIGTERM, signal.SIGINT])

    def print_stats(self, sig, stack):
        values = {
            "queue_size": self.queue.qsize(),
            "sender_processes": [p.pid for p in self.sender_processes],
            "campaign_managers": [
                p.pid for p, _ in self.campaign_manager_processes.values()
            ],
            "bucket_capacity": self.rate_limiter.peek(),
            "sending_rate": self.rate_meter.current_rate(),
        }

        self.stderr.write(self.STATS_MESSAGE % values, ending="\n")

    def start_sender_processes(self):
        for i in range(
            app_settings.MAX_CONCURRENT_SENDERS - len(self.sender_processes)
        ):
            recv_conn, send_conn = mp.Pipe(duplex=False)
            process = mp.Process(
                target=sender_process,
                kwargs={
                    "queue": self.queue,
                    "error_channel": send_conn,
                    "rate_limiter": self.rate_limiter,
                    "rate_meter": self.rate_meter,
                    "quit_event": self.senders_quit_event,
                },
            )
            process.daemon = True

            self.sender_processes.append(process)
            self.sender_pipes.append(recv_conn)
            # let's close SQL connection to make sure it is not shared with children
            connection.close()
            with self._setup_signal_handlers_for_children():
                process.start()
            logger.info(
                _("Started sender process %(process_pid)s")
                % {"process_pid": process.pid}
            )

    def check_campaigns(self):
        campaigns = Campaign.objects.filter(
            status__in=[Campaign.STATUS_WAITING, Campaign.STATUS_SENDING]
        )

        for campaign in campaigns:
            if (
                campaign.status == Campaign.STATUS_WAITING
                and campaign.id in self.campaign_manager_processes
            ):
                # we need to cancel that task
                logger.info(
                    _(
                        "Stopping campaign manager n°%(campaign_id)s (%(campaign_name)s)..."
                    )
                    % {"campaign_id": campaign.id, "campaign_name": campaign.name[20:]}
                )
                _process, quit_event = self.campaign_manager_processes[campaign.id]
                quit_event.set()

            if (
                campaign.status == Campaign.STATUS_SENDING
                and campaign.id not in self.campaign_manager_processes
            ):
                quit_event = mp.Event()
                process = mp.Process(
                    target=campaign_manager_process,
                    kwargs={
                        "campaign": campaign,
                        "queue": self.queue,
                        "quit_event": quit_event,
                    },
                )
                process.daemon = True
                self.campaign_manager_processes[campaign.id] = (process, quit_event)
                # let's close SQL connection to make sure it is not shared with children
                connection.close()
                with self._setup_signal_handlers_for_children():
                    process.start()
                logger.info(
                    _("Started campaign manager %(process_pid)s for %(campaign)s")
                    % {"process_pid": process.pid, "campaign": repr(campaign)}
                )

    def monitor_processes(self):
        sender_sentinels = [p.sentinel for p in self.sender_processes]
        campaign_manager_sentinels = {
            p.sentinel: c_id
            for c_id, (p, _e) in self.campaign_manager_processes.items()
        }

        events = mpc.wait(
            sender_sentinels + list(campaign_manager_sentinels) + self.sender_pipes,
            timeout=app_settings.POLLING_INTERVAL,
        )

        stopped_senders = []
        stopped_campaign_managers = []
        campaign_errors = set()

        for event in events:
            if event in sender_sentinels:
                stopped_senders.append(sender_sentinels.index(event))
            if event in campaign_manager_sentinels:
                stopped_campaign_managers.append(campaign_manager_sentinels[event])
            elif event in self.sender_pipes:
                while event.poll():
                    try:
                        campaign_errors.add(event.recv())
                    except EOFError:
                        break

        # reverse=True is important as we're removing elements from the sender_processes list by index
        for i in sorted(stopped_senders, reverse=True):
            process = self.sender_processes[i]
            # let's reap process to avoid zombies
            process.join()

            logger.error(
                _("Sender process %(process_pid)s unexpectedly quit...")
                % {"process_pid": process.pid}
            )
            del self.sender_processes[i]
            del self.sender_pipes[i]

        for campaign_id in sorted(stopped_campaign_managers):
            process, _quit_event = self.campaign_manager_processes[campaign_id]
            pid = process.pid
            # let's reap process to avoid zombies
            process.join()
            try:
                campaign = Campaign.objects.get(id=campaign_id)
            except Campaign.DoesNotExist:
                # the campaign has most likely be deleted
                campaign = None

            if campaign and campaign.status != campaign.STATUS_SENDING:
                # process was asked to stop and did so correctly
                logger.info(
                    _(
                        "Campaign manager %(pid)s correctly stopped... Was taking care of %(campaign)s."
                    )
                    % {"pid": pid, "campaign": repr(campaign)}
                )
            else:
                logger.error(
                    _(
                        "Campaign manager %(pid)s abruptly stops. Was taking care of %(campaign)s."
                    )
                    % {"pid": pid, "campaign": repr(campaign)}
                )

            del self.campaign_manager_processes[campaign_id]

        for campaign_id in campaign_errors:
            Campaign.objects.filter(id=campaign_id).update(status=Campaign.STATUS_ERROR)
            if campaign_id in self.campaign_manager_processes:
                logger.error(
                    _(
                        "Unexpected error while trying to send message from campaign %(campaign_id)s...\n"
                    )
                    % {"campaign_id": campaign_id}
                )
                self.campaign_manager_processes[campaign_id][1].set()

    def run_loop(self):
        try:
            while True:
                self.start_sender_processes()
                self.check_campaigns()
                self.monitor_processes()
        except (GracefulExit, KeyboardInterrupt):
            logger.info(_("Asked to quit, asking all subprocesses to exit..."))
            self.senders_quit_event.set()
            for _proces, event in self.campaign_manager_processes.values():
                event.set()

            logger.info(_("Waiting for all subprocesses to gracefully exit..."))
            # active_children joins children so removes zombies
            while mp.active_children():
                mpc.wait(
                    [p.sentinel for p in self.sender_processes]
                    + [p.sentinel for p, _e in self.campaign_manager_processes.values()]
                )

            logger.info(_("All subprocesses have exited!"))



# Generated by Django 3.1.1 on 2021-01-18 11:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("nuntius", "0019_add_tracking_index"),
    ]

    operations = [
        migrations.AlterField(
            model_name="campaign",
            name="utm_name",
            field=models.CharField(
                blank=True,
                help_text="Value used as utm_campaign parameter, used by various analytics tools.",
                max_length=255,
                verbose_name="UTM name (visible to subscribers)",
            ),
        ),
        migrations.AddIndex(
            model_name="campaignsentevent",
            index=models.Index(
                fields=["subscriber", "datetime"], name="nuntius_cam_subscri_f43488_idx"
            ),
        ),
    ]



# Generated by Django 3.1.1 on 2020-09-14 07:37

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("nuntius", "0017_campaign_utm_name")]

    operations = [migrations.RemoveField(model_name="campaign", name="task_uuid")]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nuntius',
 'nuntius.admin',
 'nuntius.management',
 'nuntius.management.commands',
 'nuntius.migrations',
 'nuntius.utils']

package_data = \
{'': ['*'],
 'nuntius': ['locale/fr/LC_MESSAGES/django.po',
             'static/nuntius/*',
             'static/nuntius/mosaico/*',
             'static/nuntius/mosaico/fontawesome/*',
             'static/nuntius/mosaico/img/*',
             'static/nuntius/mosaico/lang/*',
             'static/nuntius/mosaico/notoregular/*',
             'static/nuntius/mosaico/skins/gray-flat/*',
             'static/nuntius/mosaico/skins/gray-flat/fonts/*',
             'static/nuntius/mosaico/skins/gray-flat/img/*',
             'static/nuntius/templates/versafix-1/*',
             'static/nuntius/templates/versafix-1/edres/*',
             'static/nuntius/templates/versafix-1/img/*',
             'static/nuntius/templates/versafix-1/img/icons/*',
             'static/nuntius/templates/versafix-1/img/social_def/*',
             'templates/nuntius/*']}

install_requires = \
['Django>=2.0', 'Pillow', 'django-stdimage', 'html2text', 'tenacity']

extras_require = \
{'anymail': ['django-anymail', 'boto3'], 'mysql': ['mysqlclient']}

setup_kwargs = {
    'name': 'nuntius',
    'version': '2.2.4',
    'description': 'A newsletter app to be used with Django or standalone.',
    'long_description': '# Nuntius\n\nNuntius is a newsletter application for Django.\n\nNuntius integrates with your Django project.\nIt is very agnostic about your subscribers and subscriber lists models.\n\nIt features [Mosaico](https://mosaico.io/), a drag-and-drop email\neditor, for sending beautiful emails to your subscribers.\n\n## How it works\n\nNuntius is agnostic about your subscribers model. You can use your current\nuse model, as long as it implements a few required methods.\n\nTo allow your end-users to choose recipients, it is your choice to implement\none or more "segment" models. Segment models implement a required method\n`get_subscribers_queryset`.\n\nYou can then create campaigns in the Django admin panel, and send them to\nexisting segments.\n\n## Installation\n\n1. Add "nuntius" to your INSTALLED_APPS setting like this:\n    ````python\n        INSTALLED_APPS = [\n            \'django.contrib.admin\',\n            \'django.contrib.auth\',\n            \'django.contrib.contenttypes\',\n            \'django.contrib.sessions\',\n            \'django.contrib.messages\',\n            \'django.contrib.staticfiles\',\n            ...\n            \'nuntius\',\n        ]\n    ````\n2. Include Nuntius urlconf in your project `urls.py` like this:\n    ````python\n        path(\'nuntius/\', include(\'nuntius.urls\')),\n    ````\n3. Define your subscriber model so it works with Nuntius.\n   You must inherit from\n   [`nuntius.models.AbstractSubscriber`](https://github.com/lafranceinsoumise/nuntius/blob/master/nuntius/models.py#L174)\n   and implement all the necessary methods. An easy way to do this is to use directly or to extend\n   [`BaseSubscriber`](https://github.com/lafranceinsoumise/nuntius/blob/master/nuntius/models.py#L204), but\n   you can implement the methods of `AbstractSubscriber` the way you want.\n    \n    Here are the methods you must implement :\n    \n    * `get_subscriber_status()`\n        must return one of `AbstractSubscriber.STATUS_CHOICES`. You can also simply\n        define a `subscriber_status` attribute.\n        \n    * `get_subscriber_email()`\n        must return a unique email address for the subscriber. You can also simply\n        define an `email` attribute.\n\n    * `get_subscriber_data()`\n        must return the dictionnary of values which can be used as substitution in\n        the emails. Default is `{"email": self.get_subscriber_email()}`.\n \n4. Tell Nuntius how to find your subscriber model in `settings.py`\n    ````python\n    NUNTIUS_SUBSCRIBER_MODEL = \'myapp.MySubscriberModel\'\n    ````\n\n5. Launch the nuntius worker in the background. In a production setting, this should be done through\n   a process monitor like upstart or systemd.\n    ```shell script\n    export DJANGO_SETTINGS_MODULE=myapp.settings\n    python ./manage.py nuntius_worker\n    ```\n\n6.  Unless you are using a custom admin site, admin panels for Nuntius will be\n[autodiscovered](https://docs.djangoproject.com/en/2.0/ref/contrib/admin/#discovery-of-admin-files)\nand added to you admin site. If you use a custom admin site, you need to register\nNuntius models with something like:\n\n    ```python\n    admin_site.register(nuntius.models.Campaign, nuntius.admin.CampaignAdmin)\n    admin_site.register(nuntius.models.CampaignSentEvent, nuntius.admin.CampaignSentEventAdmin)\n    ```\n\n## Other settings\nUse `NUNTIUS_DEFAULT_FROM_EMAIL`, `NUNTIUS_DEFAULT_FROM_NAME`, `NUNTIUS_DEFAULT_REPLY_TO_EMAIL`,\n`NUNTIUS_DEFAULT_REPLY_TO_NAME` to change default field values in the admin form.\n\n## Advanced usage\n\n### List segments\n\nIf you want to have more control on your recipients, you can create a\nsegment model.\n\nOne example of segment is a simple model which holds a Many-to-Many relation\nto subscribers.\n\nAnother example is a segment model which filters subscribers depending on\nthe date of their last login :\n\n```python\nfrom django.db import models\nfrom django.db.models import fields\nfrom datetime import datetime\n\nfrom nuntius.models import BaseSegment\n\n\nclass LastLoginDateSegment(BaseSegment, models.Model):\n     last_login_duration = fields.DurationField()\n     \n     def get_display_name(self):\n         return f\'Last login : {str(datetime.now() - self.last_login_duration)}\'\n         \n     def get_subscribers_queryset(self):\n        return MySubscriberClass.objects.filter(last_login__gt=datetime.now() - self.last_login_duration)\n        \n     def get_subscribers_count(self):\n        return MySubscriberClass.objects.filter(last_login__gt=datetime.now() - self.last_login_duration, subscribed=True)\n\n```\n \n* `get_subscribers_queryset` is allowed to return subscribers regardless of their\n    `subscriber_status`, as `get_subscriber_status` will be called on each instance.\n* `get_subscribers_count` is only there for convenience in the admin panel, it does not\n    have to be accurate. If you want to have it accurate, you should however take\n    your subscribers status into account.\n\nThen, add your segment model to Nuntius settings :\n````python\nNUNTIUS_SEGMENT_MODEL = \'myapp.lastlogindatesegment\'\n````\n\n### Custom template\n\nYou can write your own Mosaico template to fit your needs. To make it available in the admin,\nlist the public URL path of the template in `NUNTIUS_MOSAICO_TEMPLATES`. The template can be served\nby Django static files system, or any other way at your preference.\n\n```python\nNUNTIUS_MOSAICO_TEMPLATES = [\n    ("/static/mosaico_templates/versafix-2/template-versafix-2.html", "Custom template")\n]\n```\n\n### Sending parameters\n\nThe worker will spawn several subprocesses to speed up the sending of campaigns. The number of\nprocesses that will send emails concurrently can be configured using the `NUNTIUS_MAX_CONCURRENT_SENDERS`\nsetting.\n\nMost ESP enforce a maximum send rate. Nuntius won\'t sent messages faster than`NUNTIUS_MAX_SENDING_RATE`,\nin messages per second.\n\nWhen using SMTP, some ESP limit the number of emails that can be sent using a single connection.\n`NUNTIUS_MAX_MESSAGES_PER_CONNECTION` will force Nuntius to reset the connection after sending that\nmany messages.\n\nThe Nuntius worker checks every `NUNTIUS_POLLING_INTERVAL` seconds if any sending has been scheduled\nor canceled. The default value of 2 seconds should be find for most usages.\n\nTo help you configure these parameters, you can send SIGUSR1 to the main worker process and it will\nprint sending statistics on `stderr`. Pay special attention to the current sending rate and to the\ncurrent bucket capacity: if your sending rate is lower than the maximum you configured, it most\nlikely means the value you chose for `NUNTIUS_MAX_CONCURRENT_SENDERS` is not high enough given\nthe latency you\'re getting with your ESP.\n\n### ESP and Webhooks\n\nMaintaining your own SMTP server to send your newsletter is probably\na bad idea if you have more than a few subscribers. You can use\n[Anymail](https://github.com/anymail/django-anymail) along with Nuntius\nin order to use an email service provider. Anymail supports\na lot of ESP, like Amazon SES, Mailgun, Mailjet, Postmark, SendGrid,\nSendinBlue, or SparkPost.\n\nRefer to the steps in [Anymail 1-2-3](https://anymail.readthedocs.io/en/stable/quickstart/)\nto install Anymail. If you want to configure Anymail just for Nuntius and keep\nthe default email backend for other usage, you can use the setting `NUNTIUS_EMAIL_BACKEND`\nrather than the default `EMAIL_BACKEND`.\n\nIn addition, configuring Nuntius with Anymail will allow you to use ESP tracking features\nand to track status of your email once it is sent.\n\n#### Webhooks\n\nConfiguring webhhoks allows Nuntius to track email status and to\ngive you statistics on campaign, as well as updating subscriber status\nwhen they bounce.\n\n1. Configure email tracking as described in\n[Anymail documentation](https://anymail.readthedocs.io/en/stable/installation/#configuring-tracking-and-inbound-webhooks).\n2. Implement the method `set_subscriber_status(self, email, status)` on your subscriber\nmodel manager.\n\nNuntius will automatically listen to Anymail signals and call this method if needed.\n\n##### Handling of non-nuntius events (optional)\n\nIf you send emails to your subscribers by other means than Nuntius (for example,\ntransactional emails), you will receive webhooks events which are not related to\na campaign you sent. By default, Nuntius will create a campaign result event recording\nthe email and the event type, but it will not link it to a campaign nor to a subscriber\nmodel.\n\nIf you want your events to always be linked to a subscriber model, you must implement\na `get_subscriber(self, email_address)` method on your subscriber model manager.\n\n\n##### BaseSubscriberManager\n\nNuntius is packaged with a BaseSubscriberManager, which implements both\n`set_subscriber_status` and `get_subscriber`, assuming you have an `email` field\non your subscriber model. This is the default manager used by `BaseSubscriber`.\n\n\n#### Bounce handling\n\nMost ESP gives you a reputation based on your hard bounce rate.\nMosaico handles bounces smartly to change your subscribers status\nwhen necessary.\n\nIf Nuntius receive a bounce event on an email address which has no\nother sending event, `set_subscriber_status(email, status)` is called\nwith `AbstractSubscriber.STATUS_BOUNCED`.\n\nIf a successful sending event exists for this address,\nthree parameters are taken into account :\n* if during the last `duration` days, there has been no more bounces than `limit`\nand at least one successful sending, no action is taken\n* if there has been at least one successful sending in the last\n`consecutive` events, no action is taken\n* otherwise, `set_subscriber_status(email, status)` is called\nwith `AbstractSubscriber.STATUS_BOUNCED`\n\n\nYou can change thoses default values :\n```python\nNUNTIUS_BOUNCE_PARAMS = {\n    "consecutive": 1,\n    "duration": 7,\n    "limit": 3\n}\n```\n\n**Example :**\n\n* You send 3 campaigns a week. After a few months, a subscriber\nhas a full mailbox. On first and second bounced campaign, no action\nis taken because there is a successful sending in the last 7 days,\nand no more than 3 bounces. On the third campaign, if the user has empty\ntheir mailbox, everything is fine. Otherwise, the subscriber is marked\nas permanently bounced.\n* You send one campaign a day. A user has a buggy email server.\nThis week, the user has already 3 bounces. When you receive the 4th\nbounce, if there has been a successful sending just before,\neverything is fine. Otherwise, the subscriber is marked\nas permanently bounced.\n\n## Tracking\n\nOpening and clicks are tracked by adding a white pixel and replacing links in emails.\n\nNuntius also adds [UTM parameters](https://en.wikipedia.org/wiki/UTM_parameters) to every URL with the following values:\n* `utm_source`: *"nuntius"*\n* `utm_medium`: *"email"*\n* `utm_campaign`: value configured by user at the campaign level\n* `utm_content`: *"link-{number}"* based on the link position in the email\n* `utm_term`: attribute `utm_term` of the segment object, or empty string if attribute does not exist\n\nIn some situations, two details may be important for you:\n\n1. `utm_campaign`, `utm_content`, and `utm_term`, those are just defaults values, and can also be set directly on\nthe link. `utm_source` and `utm_medium` will always be overwritten.\n2. `utm_content` and `utm_term` are set at sending time and cannot change afterwards. `utm_campaign` is\nset at click time, during the redirection from nuntius tracking URL to target URL, so if you change the value\nat the campaign level after sending, the value will change for all new clicks.\n\n## License\n\nCopyright is owned by Jill Royer and Arthur Cheysson.\n\nYou can use Nuntius under GPLv3 terms.\n',
    'author': 'Jill Royer',
    'author_email': 'hello@jillroyer.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lafranceinsoumise/nuntius',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# Generated by Django 2.2.5 on 2020-01-28 09:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("nuntius", "0016_indices_and_ordering")]

    operations = [
        migrations.AddField(
            model_name="campaign",
            name="utm_name",
            field=models.CharField(
                blank=True,
                help_text="Name used as utm_campaign parameter, used by various analytics tools.",
                max_length=255,
                verbose_name="UTM name (visible to subscribers)",
            ),
        )
    ]
